# -*- coding: utf-8 -*-
import os, jinja2
import numpy as np
import scipy.optimize

from util import geometry as g
from util import tools, constants

# see README for terminology, terminolology, lol
class Vertex():
    """ point with an index that's used in block and face definition
    and can output in OpenFOAM format """
    def __init__(self, point):
        self.point = np.array(point)
        self.mesh_index = None # will be changed in Mesh.prepare_data()
        
    def __repr__(self):
        s = constants.vector_format(self.point)
        
        if self.mesh_index is not None:
            s += " // {}".format(self.mesh_index)
        
        return s

class Edge():
    def __init__(self, index_1, index_2, points):
        """ an edge is defined by two vertices and points in between;
        a single point edge is treated as 'arc', more points are
        treated as 'spline'.

        passed indexes refer to position in Block.edges[] list; Mesh.prepare_data()
        will assign actual Vertex objects.
        """

        # indexes in block.edges[] list
        self.block_index_1 = index_1
        self.block_index_2 = index_2

        # these will refer to actual Vertex objects after Mesh.prepare_data()
        self.vertex_1 = None
        self.vertex_2 = None

        self.type, self.points = self.get_type(points)

    @staticmethod
    def get_type(points):
        """ returns edge type and a list of points:
        'None' for a straight line,
        'arc' for a circular arc,
        'spline' for a spline """

        if points is None:
            return None, None

        # if multiple points are given check that they are of correct length
        points = np.array(points)
        shape = np.shape(points)

        if len(shape) == 1:
            t = 'arc'
        else:
            assert len(shape) == 2
            for p in points:
                assert len(p) == 3
            
            t = 'spline'

        return t, points
        

    @property
    def point_list(self):
        if self.type == 'arc':
            return constants.vector_format(self.points)
        else:
            return "(" +  \
                   " ".join([constants.vector_format(p) for p in self.points]) + \
                   ")"
    
    @property
    def is_valid(self):
        # 'all' spline edges are 'valid'
        if self.type == 'spline':
            return True
        
        # wedge geometries produce coincident 
        # edges and vertices; drop those
        if g.norm(self.vertex_1.point - self.vertex_2.point) < constants.tol:
            return False

        # if case vertex1, vertex2 and point in between
        # are collinear, blockMesh will find an arc with
        # infinite radius and crash.
        # so, check for collinearity; if the three points
        # are actually collinear, this edge is redundant and can be
        # silently dropped
        OA = self.vertex_1.point
        OB = self.vertex_2.point
        OC = self.points

        # if point C is on the same line as A and B:
        # OC = OA + k*(OB-OA)
        AB = OB - OA
        AC = OC - OA

        k = g.norm(AC)/g.norm(AB)
        d = g.norm((OA+AC) - (OA + k*AB))

        return d > constants.tol

    def get_length(self):
        # TODO: test
        def curve_length(points):
            l = 0

            for i in range(len(points)-1):
                l += g.norm(points[i+1] - points[i])

            return l

        if self.type == 'arc':
            edge_points = np.array([
                self.vertex_1.point,
                self.points,
                self.vertex_2.point
            ])

            return curve_length(edge_points)
        elif self.type == 'spline':
            print(self.vertex_1.point)
            print(self.points)
            print(self.vertex_2.point)

            edge_points = np.concatenate((
                [self.vertex_1.point],
                self.points,
                [self.vertex_2.point]), axis=0)
            print("edges: ", edge_points)
            return curve_length(edge_points)
        else:
            raise AttributeError(f"Unknown edge type: {self.type}")

    def __repr__(self):
        return "{} {} {} {}".format(
            self.type,
            self.vertex_1.mesh_index,
            self.vertex_2.mesh_index,
            self.point_list
        )