#!/usr/bin/env python
import sys
import os
import math

import numpy as np

from classes.primitives import Edge
from classes.mesh import Mesh
from operations.base import Face
from operations.operations import Loft
from util import geometry as g

import parameters as p

#####################################
# generate a circular arc of specified thickness in x-z plane
# by rotating points on x-axis around x-axis back and forth
def rot(point, angle):
    # rotates a 'point' by an 'angle' around z-axis
    return g.arbitrary_rotation(point, [0, 0, 1], angle, [0, 0, 0])

def circular_face(r, beta, arc):
    # creates a circular Face on specified 'r',
    # starting from angle 'beta' and spanning 'arc' angle 

    # r - radial location of the created face
    internal_vertex = np.array([p.pipe_diameter/2 - p.pipe_thickness, 0, r])
    external_vertex = np.array([p.pipe_diameter/2,                    0, r])

    inner_vertices = [
        rot(internal_vertex, beta),
        rot(internal_vertex, beta + arc),
        rot(external_vertex, beta + arc),
        rot(external_vertex, beta)
    ]

    # add a single point so that blockMesh will generate a circular edge
    inner_edges = [
        rot(internal_vertex, beta + arc/2), None,
        rot(external_vertex, beta + arc/2), None
    ]

    return Face(inner_vertices, inner_edges)

# generate this many blocks span-wise;
# the higher the number, the smoother the curve
n_blocks = 10

mesh = Mesh()
betas = np.linspace(p.beta_inner, p.beta_outer, num=n_blocks+1)
arcs = np.linspace(p.arc_inner, p.arc_outer, num=n_blocks+1)
radii = np.linspace(p.r_inner, p.r_outer, num=n_blocks+1)

for i in range(n_blocks):
    inner_face = circular_face(radii[i], betas[i], arcs[i])
    outer_face = circular_face(radii[i+1], betas[i+1], arcs[i+1])

    block = Loft(outer_face, inner_face)
    block.count_to_size(0, p.pipe_thickness/2)
    block.set_cell_count(1, 1)
    block.set_cell_count(2, 10)

    mesh.add_operation(block)

mesh.write('util/blockMeshDict.template', 'blockMeshDict.blade')
print("blockMeshDict.blade created")